const util = require('util');
const lodash = require('lodash');
const { Gauge, Counter, Summary } = require('prome-string');

class Prometheus {
  constructor({ ...tags } = {}) {
    this.tags = lodash.pickBy(tags, (v) => v !== undefined);
    const KEYS = Object.keys(this.tags);

    this.table = {
      gauge: new Gauge({
        name: 'gauge',
        help: '',
        labels: [...KEYS, 'name'],
      }),

      callCounter: new Counter({
        name: 'call_counter',
        help: '',
        labels: [...KEYS, 'module', 'method'],
      }),

      callDuration: new Summary({
        name: 'call_duration',
        help: '',
        labels: [...KEYS, 'module', 'method'],
        queueLength: 1000,
        percentiles: [0.5, 0.8, 0.9, 0.99],
        timeout: 24 * 3600 * 1000,
      }),
    };
  }

  setGauge(name, count) {
    if (count !== undefined) {
      this.table.gauge.set(count, { ...this.tags, name });
    }
  }

  toString() {
    return Object.values(this.table).join('\n');
  }

  // ---------------------------------------------------------------------------
  decorate(func, options = {}) {
    const prometheus = this;
    return async function (...args) {
      const label = { ...prometheus.tags, ...options };

      const timestamp = Date.now();
      try {
        return await func.call(this, ...args);
      } finally {
        prometheus.table.callCounter.add(1, label);
        prometheus.table.callDuration.set(Date.now() - timestamp, label);
      }
    };
  }

  decorateMethod(object, method, options = {}) {
    const func = lodash.get(object, method);
    if (!lodash.isFunction(func)) {
      return;
    }

    options.module = options.module || object.constructor.name;

    object[method] = this.decorate(func, { module, method, ...options });
  }

  decorateObject(object, options = {}) {
    if (!lodash.isObject(object)) {
      return;
    }

    const descriptors = Object.getOwnPropertyDescriptors(object);
    lodash.forEach(descriptors, (descriptor, name) => {
      // XXX: only `public` and `async` function to be trace
      if (!name.startsWith('_') && util.types.isAsyncFunction(descriptor.value)) {
        this.decorateMethod(object, name, options);
      }
    });
  }

  decorateModule(instance, options = {}) {
    if (!lodash.isObject(instance)) {
      return;
    }

    const object = Object.getPrototypeOf(instance);
    const descriptors = Object.getOwnPropertyDescriptors(object);
    const nameArray = Object.keys({ ...descriptors, ...instance });
    for (const name of nameArray) {
      const descriptor = Object.getOwnPropertyDescriptor(object, name);
      if (!name.startsWith('_') && descriptor && util.types.isAsyncFunction(descriptor.value)) {
        this.decorateMethod(instance, name, options);
      }
    }
  }
}

module.exports = Prometheus;
